/**
 * envfs - (c) Andrew Flegg 2004. Released under the GPL.
 * ~~~~~   mailto:andrew@bleb.org
 *
 * Based on:
 *     http://www.geocities.com/ravikiran_uvs/articles/envfs.html
 *     /usr/src/linux-2.6.6/Documentation/filesystems/porting
 * and http://lwn.net/Articles/driver-porting/
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/statfs.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/namei.h>
#include <linux/list.h>

#define ENVFS_MAGIC 0xabcd
#define FILE_INODE_NUMBER 2


/*
 * We don't need no steenkin' .h file...
 */
char *env_get(char *);
static int envfs_fill_super(struct super_block *, void *, int);
struct super_block *envfs_get_sb(struct file_system_type *, int,
				 const char *, void *);
void envfs_kill_super(struct super_block *);
void envfs_s_readinode(struct inode *);
int envfs_s_statfs(struct super_block *, struct kstatfs *);
int envfs_i_readlink(struct dentry *, char __user *, int);
int envfs_i_follow_link(struct dentry *, struct nameidata *);
struct dentry *envfs_i_lookup(struct inode *, struct dentry *, struct nameidata *);
int envfs_f_readdir(struct file *, void *, filldir_t);


/*
 * Data declarations
 */
static struct super_operations envfs_sops = {
    read_inode: envfs_s_readinode,
    statfs: envfs_s_statfs
};

struct inode_operations envfs_dir_iops = {
    lookup: envfs_i_lookup,
};

struct inode_operations envfs_link_iops = {
    readlink: &envfs_i_readlink,
    follow_link: &envfs_i_follow_link
};

struct file_operations envfs_dir_fops = {
    read   : generic_read_dir,
    readdir: &envfs_f_readdir
};


static struct file_system_type envfs = {
	.owner		= THIS_MODULE,
	.name		= "envfs",
	.get_sb		= envfs_get_sb,
	.kill_sb	= envfs_kill_super,
	.fs_flags	= 0
};

struct inode *envfs_root_inode;


/**
 * Expand an environment variable. Only return vars with values which start
 * with '/' (ie. absolute paths). Also, if value contains several paths separated
 * by ':', only return the first entry.
 */
char *env_get(char *var) {
    char *value = NULL;

    struct mm_struct *mm;
    task_lock(current);
    if ((mm = current->mm))
        atomic_inc(&mm->mm_users);
    task_unlock(current);

    if (mm) {
        char *env = (char *)mm->env_start;
	size_t s  = strlen(var);
        while (env < (char*) mm->env_end) {
            if (!(strncmp(env, var, s)) && env[s] == '=' && env[s+1] == '/') {
		value = env + s + 1;
		char *end = strchr(value, ':');
		if (end != NULL) end[0] = 0;
		break;
	    }
	    env = strchr(env, 0) + 1;
	}
	atomic_dec(&mm->mm_users);
    }

    return value;
}
    

/* -- Filesystem Operations --------------------------------------------- */
/**
 * This will be called when the kernel is attempting to mount something. It
 * creates a super_block structure and calls our callback/function to fill it.
 */
static int envfs_fill_super( struct super_block *sb, void *buf, int size ) {
    sb->s_blocksize = 1024;
    sb->s_blocksize_bits = 10;
    sb->s_magic = ENVFS_MAGIC;
    sb->s_op = &envfs_sops;                        // super block operations
    sb->s_type = &envfs;                           // file_system_type

    envfs_root_inode = iget( sb, 1 );              // allocate an inode
    envfs_root_inode->i_op = &envfs_dir_iops;      // set the inode ops
    envfs_root_inode->i_mode = S_IFDIR|0555; 
    envfs_root_inode->i_fop = &envfs_dir_fops;     // set the directory ops

    if(!(sb->s_root = d_alloc_root(envfs_root_inode))) {
        iput(envfs_root_inode);
        return -EINVAL;
    }

    printk( "envfs: fill_super succeeded\n" );
    return 0;
}


/**
 * Wrapper the fill_super call. Required as part of the 2.6 kernel interface.
 */
struct super_block *envfs_get_sb(struct file_system_type *fs_type,
					int flags, const char *dev_name,
					void *data) {
    return get_sb_nodev(fs_type, flags, data, envfs_fill_super);
}


/**
 * Filesystem's going away.
 */
void envfs_kill_super(struct super_block *sb) {
    printk("envfs: envfs_kill_super... ");
    kill_litter_super(sb);
    printk("done.\n");
}


/* -- Super Block Operations -------------------------------------------- */
/**
 * Something's attempting to read an inode, set the modification times to
 * the current time.
 */
void envfs_s_readinode( struct inode *inode ) {
    inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
    printk( "envfs: super_operations.readinode called on %d\n", (int)inode->i_ino );
}


/**
 * This will be called to get the filesystem information like disk usage etc.
 */
int envfs_s_statfs( struct super_block *sb, struct kstatfs *buf ) {
    buf->f_type = ENVFS_MAGIC;
    buf->f_bsize = PAGE_SIZE/sizeof(long);
    buf->f_bfree = 0;
    buf->f_bavail = 0;
    buf->f_ffree = 0;
    buf->f_namelen = NAME_MAX;
    printk( "envfs: super_operations.statfs called\n" );
    return 0;
}


/* -- Inode Operations -------------------------------------------------- */
/**
 * Get the destination of a link. Called by things such as `ls'.
 *
 * @param dentry Directory entry of symlink.
 * @param buffer Destination buffer for string containing symlink's target.
 * @param buflen Size of destination buffer.
 * @return Length of string for symlink's target.
 */
int envfs_i_readlink( struct dentry *dentry,
 	 	      char __user *buffer,
		      int buflen) {
    char *name = (char *)dentry->d_name.name;
    printk("envfs: readlink %s\n", name);

    char *dest = env_get(name);
    int  len   = strlen(dest);
    if (buflen < len) len = buflen;

    strncpy(buffer, dest, len);
    printk("envfs: %s -> %s\n", name, dest);
    return len;
}


/**
 * Fill a nameidata structure for the target of a symlink. Used when
 * the symlink is actually followed.
 */
int envfs_i_follow_link( struct dentry *dentry,
			 struct nameidata *nd) {
    char *name = (char *)dentry->d_name.name;
    printk("envfs: follow_link %s\n", name);

    char *dest = env_get(name);
    /*return path_lookup(dest, LOOKUP_FOLLOW, nd); */
    return vfs_follow_link(nd, dest);
}


/**
 * Called to determine if an object actually exists. This is the
 * initial check and fills in data for things such as file size,
 * mode etc.
 *
 * @param parent_inode Inode of containing directory.
 * @param dentry Directory entry of object being looked up.
 * @param nd Nameidata (TODO: being passed in or being passed out?)
 * @return Pointer to error object, or NULL (TODO: why not int?!)
 */
struct dentry *envfs_i_lookup( struct inode *parent_inode,
			       struct dentry *dentry,
			       struct nameidata *nd) {
    struct inode *file_inode;
    char *path = (char*)dentry->d_name.name;
    char *value = "unknown";
    int len = 0;
    printk("envfs: lookup on %s\n", path);
 
    // Is this not us?
    if( parent_inode->i_ino != envfs_root_inode->i_ino )
        return ERR_PTR(-ENOENT);

    if (!strcmp(path, "/"))
	return NULL;

    // Allocate an inode object
    if(!(file_inode = iget( parent_inode->i_sb, FILE_INODE_NUMBER )))
        return ERR_PTR(-ENOMEM);

    value = env_get(path);
    if ((value == NULL) || ((len = strlen(value)) == 0))
        return ERR_PTR(-ENOENT);
  
    file_inode->i_size = len;
    file_inode->i_mode = S_IFLNK | 0777;
    file_inode->i_op  = &envfs_link_iops;
    // file_inode->i_fop = &envfs_file_fops;

    // Add the inode to the dentry object
    d_add(dentry, file_inode);
    printk("envfs: inode_operations.lookup called for %s, size = %d, value = %s\n", path, len, value);
    return NULL;
}


/* -- File/Directory Operations ----------------------------------------- */
/**
 * Read a directory's contents.
 */
int envfs_f_readdir( struct file *file, void *dirent, filldir_t filldir ) {
    struct dentry *de = file->f_dentry;

    printk( "envfs: file_operations.readdir called on %s\n", de->d_name.name );

    if(file->f_pos > 0 )
        return 1;
    if(filldir(dirent, ".", 1, file->f_pos++, de->d_inode->i_ino, DT_DIR)||
       (filldir(dirent, "..", 2, file->f_pos++, de->d_parent->d_inode->i_ino, DT_DIR)))
        return 0;

    return 1;
}


/* -- Module initialisation cruft --------------------------------------- */
int init_module(void) {
    int err;
    printk("envfs: registering filesystem... ");
    err = register_filesystem( &envfs );
    if (err)
	printk("failed (error=%d)\n", err);
    else
        printk("done.\n");
    
    return err;
}

void cleanup_module(void) {
    int err;
    
    printk("envfs: unregistering filesystem... ");
    err = unregister_filesystem( &envfs );
    if (err)
	printk("failed (error=%d)\n", err);
    else
        printk("done.\n");

    return;
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andrew Flegg");


